"""
A module for plotting
"""

# Imports
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import pdist, cdist, squareform
from scipy import constants

from .utils import masses

# Constants
FONT_SIZE = 8
LINE_WIDTH = 0.5
EVCONV = 1e10 * constants.e / (4 * np.pi * constants.epsilon_0)


def init_matplotlib():
    """
    Initializes matplotlib settings
    """

    plt.rc(
        "font",
        family="serif",
        serif="Times New Roman",
        size=FONT_SIZE,
    )

    plt.rc(
        "mathtext",
        fontset="cm",
        )

    plt.rc(
        "axes",
        linewidth=LINE_WIDTH,
        labelsize=FONT_SIZE,
    )

    plt.rc(
        "lines",
        linewidth=LINE_WIDTH,
    )

    tickparams = {
            'major.width': LINE_WIDTH,
            'minor.width': LINE_WIDTH,
            'direction': 'in',
    }

    plt.rc('xtick', **tickparams)
    plt.rc('ytick', **tickparams)


def evaluate_potential(coords: np.ndarray,
                       q: np.ndarray,
                       num_points: int,
                       N: int):

    """
    Evalulates the potential on a 2D grid in the xy plane

    Parameters
    ----------
    coords : np.ndarray
        The coordinates of the exact charge region and 
        the parameter region
    q : np.ndarray
        The charges of the exact charge region and 
        the parameter region
    num_points : int
        The number of points in the grid
    N : int
        The dimension of the grid

    Returns
    -------
    x : np.ndarray
        The x coordinates of the grid
    y : np.ndarray
        The y coordinates of the grid
    z : np.ndarray
        The potential on the grid
    """

    data = np.hstack((coords, q.reshape(-1, 1)))

    x = np.linspace(-N[0], N[0], num_points)
    y = np.linspace(-N[1], N[1], num_points)
    x, y = np.meshgrid(x, y)
    z = np.zeros_like(x)

    for i in range(len(x)):
        for j in range(len(x)):
            for k in range(len(data)):
                z[i, j] += data[k, 3] / np.sqrt((x[i, j] - data[k, 0])**2 +
                                                (y[i, j] - data[k, 1])**2 +
                                                (data[k, 2])**2)

    return x, y, z


def plot_potential(x: np.ndarray,
                   y: np.ndarray,
                   z: np.ndarray,
                   num_contours: int,
                   show: bool,
                   ):

    """
    Plots the potential on a 2D grid in the xy plane

    Parameters
    ----------
    x : np.ndarray
        The x coordinates of the grid
    y : np.ndarray
        The y coordinates of the grid
    z : np.ndarray
        The potential on the grid
    num_contours : int
        The number of contours to plot
    show : bool
        Whether to show the plot or not
    """

    # Initialize matplotlib settings
    init_matplotlib()

    # Plot the potential
    fig, ax = plt.subplots(figsize=(4.5, 4.5))
    ax.contourf(x, y, z, num_contours, cmap='viridis')
    ax.set_xlabel('x/ Å')
    ax.set_ylabel('y/ Å')
    ax.set_aspect('equal')
    cbar = fig.colorbar(ax.contourf(x, y, z, num_contours, cmap='viridis'))
    cbar.set_label('Potential/ eV')
    fig.savefig('potential.png', dpi=500)
    if show:
        plt.show()


def plot_3d_model(sphere,
                  param,
                  qm,
                  ):
    """
    Plots the 3D model of the system

    Parameters
    ----------
    sphere : np.ndarray
        The exact charge region
    param : np.ndarray
        The parameter region
    qm : np.ndarray
        The quantum mechanical region
    """

    init_matplotlib()

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    # plot the coordiates as a 3d scatter
    ax.scatter(sphere[:, 0], sphere[:, 1], sphere[:, 2], c='r',
               marker='o', alpha=0.12, s=3)
    ax.scatter(param[:, 0], param[:, 1], param[:, 2], c='b',
               marker='o', alpha=0.12, s=3)
    ax.scatter(qm[:, 0], qm[:, 1], qm[:, 2], c='red', marker='o',
               alpha=1, s=5)

    ax.set_xlabel('X / Å')
    ax.set_ylabel('Y / Å')
    ax.set_zlabel('Z / Å')
    ax.legend(['Exact charge region', 'Paramamter region', 'Quantum mechanical region'])
    ax.set_aspect('equal')

    fig.tight_layout()

    # draw onto canvas
    plt.show()



"""
Computed the infinite periodic electrostatic potential
using the Ewald method and fits a set of arbitrary point
charges to the potential.
"""

import numpy as np
from scipy import constants
from scipy.special import erfc
from scipy.spatial.distance import cdist, pdist, squareform
from scipy.optimize import lsq_linear
from time import perf_counter

from .structure import Parser, generate_qm_idc
from .output import OutputWriter
from .cells import UnitSupercell, UnitCluster


###################
# Miguel was here #
###################

ALPHA = 0.2
EVCONV = 1e10 * constants.e / (4 * np.pi * constants.epsilon_0)


def create_images(coords: np.ndarray,
                  n: np.ndarray,  # Summation limits
                  lattice_vectors: np.ndarray
                  ) -> np.ndarray:
    """
    Creates the periodic images of the unit cell of dimensions
    of the summation limits.

    Parameters
    ----------
    coords : np.ndarray
        The coordinates of the unit cell.
    n : np.ndarray
        The summation limits.
    lattice_vectors : np.ndarray
        The lattice vectors of the unit cell.

    Returns
    -------
    images : np.ndarray
        The periodic images of the unit cell.
    """

    n1 = list(range(-n[0], n[0]+1))
    n2 = list(range(-n[1], n[1]+1))
    n3 = list(range(-n[2], n[2]+1))
    _n = np.array([[n1[i], n2[j], n3[k]]
                   for i in range(len(n1))
                   for j in range(len(n2))
                   for k in range(len(n3))])

    images = np.array([coord + _n[i]
                       for i in range(len(_n))
                       for coord in coords])

    return images @ lattice_vectors, _n


def calculate_fm(n, bv):
    """
    Calculates the reciprocal lattice vector images

    Parameters
    ----------
    n : np.ndarray
        The summation image indecies.
    bv : np.ndarray
        The reciprocal lattice vectors of the unit cell.

    Returns
    -------
    fm : np.ndarray
        The reciproal lattice vector images.
    """
    m = [x for x in n if not (x == 0).any()]
    return np.array(list(map(lambda x: bv @ x, m)), dtype=float)


def reciprocal_exponential(fm):
    """
    Calculates the reciprocal exponential term of the Ewald sum.

    Parameters
    ----------
    fm : np.ndarray
        The reciprocal lattice vector image.

    Returns
    -------
    reciprocal_exponential : float
    """
    return np.exp(-(np.pi**2 * fm @ fm / ALPHA**2)) / (fm @ fm)


def reciprocal_cosine(fm, r1, r2):
    """
    Calculates the reciprocal cosine term of the Ewald sum.

    Parameters
    ----------
    fm : np.ndarray
        The reciprocal lattice vector image.
    r1 : np.ndarray
        The coordinate of atom i.
    r2 : np.ndarray
        The coordinate of atom j.

    Returns
    -------
    reciprocal_cosine : float
    """
    if (r1 == r2).all():
        return 0
    else:
        return np.cos(2 * np.pi * fm @ (r1 - r2))


def ewald_reciprocal(coords: np.ndarray,
                     lattice_vectors: np.ndarray,
                     reciprocal_vectors: np.ndarray,
                     q: np.ndarray,
                     n: np.ndarray,  # Summation image indecies
                     volume: float,
                     ) -> np.ndarray:
    """
    Calculates the reciprocal space contribution to the potential

    Parameters
    ----------
    coords : np.ndarray
        The coordinates of the unit cell.
    lattice_vectors : np.ndarray
        The lattice vectors of the unit cell.
    reciprocal_vectors : np.ndarray
        The reciprocal lattice vectors of the unit cell.
    q : np.ndarray
        The charges of the atoms in the unit cell.
    n : np.ndarray
        The summation image indecies.
    volume : float
        The volume of the unit cell.

    Returns
    -------
    V_recip : np.ndarray
    """
    coords = coords @ lattice_vectors
    prefactor = EVCONV / (np.pi * volume)
    fm = calculate_fm(n, reciprocal_vectors)
    expo = np.array(sum(list(map(reciprocal_exponential, fm))))
    cosine = np.array(list(map(lambda r: sum(list(map(lambda x, y: reciprocal_cosine(x, r, y), fm, coords))), coords)), dtype=float)
    return prefactor * q * expo * cosine


def error_function(q, r1, r2):
    """
    Calculates the complementary error function term of the Ewald sum.

    Parameters
    ----------
    q : float
        The charge of atom j.
    r1 : np.ndarray
        The coordinate of atom i.
    r2 : np.ndarray
        The coordinate of atom j.

    Returns
    -------
    error_function : float
    """
    if (r1 == r2).all():
        return 0
    else:
        return q * erfc(ALPHA * np.linalg.norm(r1 - r2))/np.linalg.norm(r1 - r2)


def ewald_real_space(coords: np.ndarray,
                     q: np.ndarray,
                     lattice_vectors: np.ndarray,
                     images: np.ndarray,
                     n: np.ndarray,
                     ):
    """
    Calculates the real space contribution to the potential

    Parameters
    ----------
    coords : np.ndarray
        The coordinates of the unit cell.
    q : np.ndarray
        The charges of the atoms in the unit cell.
    lattice_vectors : np.ndarray
        The lattice vectors of the unit cell.
    images : np.ndarray
        The periodic images of the unit cell.
    n : np.ndarray
        The summation image indecies.
    """
    coords = coords @ lattice_vectors
    q = np.tile(q, len(n))
    error_component = np.array(list(map(lambda r: sum(list(map(lambda x: error_function(q[x], r, images[x]), range(len(q))))), coords)), dtype=float)
    return EVCONV * error_component


def ewald_self(q: np.ndarray,
               ) -> np.ndarray:
    """
    Computes the self interaction contribution to the Ewald sum.

    Parameters
    ----------
    q : np.ndarray

    Returns
    -------
    self : np.ndarray
    """

    prefactor = -2 * ALPHA / np.sqrt(np.pi)
    e_self = np.array(
            list(map(lambda x: prefactor * x, q))
            ).flatten()
    return e_self * EVCONV


def build_zones(cluster: UnitSupercell,
                ewald_potential: np.ndarray,
                r_cut: float,
                q: np.ndarray,
                ):
    """
    Builds the exact point charge region and the parameter
    region.

    Parameters
    ----------
    cluster : UnitSupercell
        The unit supercell.
    ewald_potential : np.ndarray
        The Ewald potential.
    r_cut : float
        The cutoff radius.
    q : np.ndarray
        The charges of the unit cell.

    Returns
    -------
    exact : np.ndarray
        The exact point charge region.
    param : np.ndarray
        The parameter region.
    """

    ewald_potential = np.tile(ewald_potential, cluster.n_cell)
    q = np.tile(q, cluster.n_cell)

    supercell = np.concatenate((cluster.cart_coords,
                                ewald_potential[:, None],
                                q[:, None]),
                               axis=1)

    sphere = np.array([coord for coord in supercell
                       if np.linalg.norm(coord[:3]) < r_cut])

    param = np.array([coord for coord in supercell
                      if np.linalg.norm(coord[:3]) > r_cut])

    return sphere, param


def drive_dipole(sphere: np.ndarray,
                 param: np.ndarray,
                 ):
    """
    Drives the total dipole moment to zero.

    Parameters
    ----------
    sphere : np.ndarray
        The exact point charge region.
    param : np.ndarray
        The parameter region.

    Returns
    -------
    sphere : np.ndarray
        The exact point charge region.
    param : np.ndarray
        The parameter region.
    """

    target = np.zeros(3)

    def calc_dipole(sphere: np.ndarray,
                    param: np.ndarray,
                    ) -> np.ndarray:
        sphere = np.array(sphere[:, :5], dtype=float)
        param = np.array(param[:, :5], dtype=float)
        d_sphere = sphere[:, 4] @ sphere[:, :3]
        d_param = param[:, 4] @ param[:, :3]
        return d_sphere + d_param

    def calc_rmsd(dipole: np.ndarray,
                  target: np.ndarray,
                  ) -> float:
        return np.sqrt(np.sum((dipole - target)**2) / len(dipole))

    initial_dipole = calc_dipole(sphere, param)

    rmsd = calc_rmsd(initial_dipole, target)
    print("Initial dipole moment: {:.3f}  {:.3f}  {:.3f}".format(
        *initial_dipole))
    print(f'Initial RMSD: {rmsd:.3f}')

    r = param[:, :3]

    d = - np.array(initial_dipole, dtype=float)
    x = np.linalg.lstsq(r.T, d, rcond=None)[0]
    param[:, 4] += x

    final_dipole = calc_dipole(sphere, param)
    rmsd = calc_rmsd(final_dipole, target)
    print("Final dipole moment: {:.3f}  {:.3f}  {:.3f}".format(
        *final_dipole))
    print(f'Final RMSD: {rmsd:.3f}')

    return param


def fit_parameters(sphere: np.ndarray,
                   param: np.ndarray,
                   verbose: int,
                   ):
    """
    Fits the parameter region such that the potential in the
    exact point charge region and the quantum mechanical (QM)
    region is the calculated Ewald potential.

    Parameters
    ----------
    sphere : np.ndarray
        The exact point charge region.
    param : np.ndarray
        The parameter region.
    verbose : int
        The verbosity level.

    Returns
    -------
    param : np.ndarray
        The parameter region.
    """

    # sum_j (q + dq) / rij - V_ewald = 0
    # sum_j (dq) / rij = V_ewald - sum_j (q) / rij
    # cx = d, d = - sum_j (q) / rij
    # QR decomposition
    # AQ = [A1(N_c), A2(N - N_c)]
    # b2 = b1 - A1(R^{-T}d)
    # Q^T x = [x1, x2]
    # A2 x2 = b2
    # x = Q @ [x1, x2]

    sphere_coords = np.array(sphere[:, :3], dtype=float)
    param_coords = np.array(param[:, :3], dtype=float)
    sphere_q = np.array(sphere[:, 4], dtype=float)
    param_q = np.array(param[:, 4], dtype=float)

    bi = np.array(sphere[:, 3], dtype=float)

    A = 1 / cdist(sphere_coords, param_coords, metric='euclidean')
    rijs = squareform(1 / pdist(sphere_coords, metric='euclidean'))
    p_sphere = rijs @ (EVCONV * sphere_q)
    p_param = A @ (EVCONV * param_q)

    bi = bi - p_sphere - p_param

    c = np.ones((1, len(param_coords)))
    d = - np.sum(sphere_q) + np.sum(param_q)

    Q, R = np.linalg.qr(c.T, mode='complete')

    AQ = A @ Q

    b2 = bi - AQ[:, 0] * (1/R[0]).T * d

    x2 = lsq_linear(AQ[:, 1:],
                    b2,
                    bounds=(-15, 15),
                    method='bvls',
                    tol=1e-6,
                    max_iter=10000,
                    verbose=verbose).x

    x1 = (1/R[0]).T * d
    x = np.concatenate((x1, x2))
    x = Q @ x

    x = x / EVCONV

    x = param_q + x

    # rmsd
    q = np.concatenate((sphere_q, x))
    coords = np.concatenate((sphere_coords, param_coords))
    rij = squareform(1 / pdist(coords, metric='euclidean'))
    rij[np.where(rij == np.inf)] = 0

    potential = EVCONV * q @ rij
    ewald = np.array(sphere[:, 3], dtype=float)
    rmsd = np.sqrt(np.sum((potential[:len(sphere)] - ewald)**2) / len(ewald))
    print(f'RMSD: {rmsd}')

    param[:, 4] = x

    return param, rmsd


def write_output(sphere: np.ndarray,
                 param: np.ndarray,
                 ):
    """
    Writes the 'ewald.out' file containing the coordinates
    and charges for using in OpenMolcas as an xfield input.

    Parameters
    ----------
    sphere : np.ndarray
        The exact point charge region.
    param : np.ndarray
        The parameter region.
    """
    with open('ewald.out', 'w') as f:
        for i in range(len(sphere)):
            f.write('{:.8f}\t{:.8f}\t{:.8f}\t{:.10f}\n'.format(
                sphere[i, 0],
                sphere[i, 1],
                sphere[i, 2],
                sphere[i, 4],
            ))
        for i in range(len(param)):
            f.write('{:.8f}\t{:.8f}\t{:.8f}\t{:.10f}\n'.format(
                param[i][0],
                param[i][1],
                param[i][2],
                param[i][4],
            ))


def check_errors(cluster, q):
    """
    Puts a limit on the number of atoms in the supercell.
    """
    coords = cluster.cart_coords
    unit_coords = cluster.unit_coords

    if len(coords) > 10000:
        print('WARNING: The number of atoms in the Ewald sum is very large. ')
        print('         This may take a long time to run.')
    elif len(coords) > 15000:
        raise ValueError('The number of atoms in the Ewald sum is too large. ')

    if len(q) != len(unit_coords):
        raise ValueError('The number of charges does not match the number of atoms.')

    if not np.allclose(np.sum(q), 0):
        raise ValueError('The total charge of the system is not zero.')


def ewald_potential(cluster: UnitCluster,
                    qm_region: np.ndarray,
                    charges: np.ndarray,
                    r_cut: int,
                    verbose: int,
                    n: list):
    """
    Calculates and fits the ewald potential

    Parameters
    ----------
    cluster : UnitCluster
        The cluster object.
    qm_region : np.ndarray
        The quantum mechanical region.
    charges : np.ndarray
        The charges of the unit cell.
    r_cut : int
        The cutoff radius.
    verbose : int
        The verbosity level.
    n : list
        The summation limits.

    Returns
    -------
    calculation dictionary: dict
        The calculation dictionary.
    """

    # Create a full calculation dict
    calc_dict = {}

    # replace with cluster ?
    av = cluster.lat_vecs
    coords = cluster.unit_coords
    bv = cluster.recip_vecs
    v = cluster.unit_volume

    # Parse charges
    q = np.loadtxt(charges)

    calc_dict.update({'Unit cell coordinates': coords,
                      'Lattice vectors': av,
                      'Reciprocal lattice vectors': bv,
                      'Unit cell volume': v,
                      'Unit cell charges': q,
                      })

    # Error check
    check_errors(cluster, q)

    images, n = create_images(coords, n, av)

    # compute the ewald sums new method
    recip = ewald_reciprocal(coords, av, bv, q, n, v)
    real = ewald_real_space(coords, q, av, images, n)
    self_int = ewald_self(q)
    potential = recip + real + self_int

    calc_dict.update({"Ewald real": real,
                      "Ewald recip": recip,
                      "Ewald self": self_int,
                      "Ewald potential": potential})

    # generate the sphere and parameter zones
    sphere, param = build_zones(cluster, potential, r_cut, q)

    # Drive the total dipole moment to zero
    param = drive_dipole(sphere, param)

    # Perform the fitting procedure and write the outputs
    param, rmsd = fit_parameters(sphere, param, verbose)

    calc_dict.update({"Fitting RMSD": rmsd})

    print('len sphere {} len param {} len qm {}'.format(len(sphere), len(param), len(qm_region)))

    # Remove qm region from sphere
    qm_idc = list(generate_qm_idc(qm_region, sphere[:, :3]))

    sphere = np.delete(sphere, qm_idc, axis=0)

    print('len sphere {} len param {} len qm {}'.format(len(sphere), len(param), len(qm_region)))


    calc_dict.update({'qm region': qm_region,
                      'Sphere': sphere,
                      'Param': param,
                      })

    return calc_dict



from itertools import product
from functools import reduce
from math import ceil
from collections import Counter

# from pymatgen.core.structure import Molecule
# from pymatgen.analysis.graphs import StructureGraph
# from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
# from pymatgen.analysis.local_env import IsayevNN
import networkx as nx
import networkx.algorithms.isomorphism as iso

from phonopy.interface.vasp import read_vasp  # TODO: replace by alternative POSCAR parser
import numpy as np
import numpy.linalg as la

from .utils import ModuleImporter


def generate_molecular_graph(structure, bonding_cutoff=3.0):

    # Imports pymatgen modules
    p_graphs = ModuleImporter('pymatgen.analysis.graphs')
    p_local_env = ModuleImporter('pymatgen.analysis.local_env')
    p_analyzer = ModuleImporter('pymatgen.symmetry.analyzer')

    # Import classes
    StructureGraph = p_graphs.StructureGraph
    IsayevNN = p_local_env.IsayevNN
    SpacegroupAnalyzer = p_analyzer.SpacegroupAnalyzer

    strategy = IsayevNN(cutoff=bonding_cutoff, allow_pathological=True)
    # Updated to use the recommended from_local_env_strategy to fix deprecation warning
    graph = StructureGraph.from_local_env_strategy(structure, strategy, weights=True)

    sym_structure = SpacegroupAnalyzer(structure).get_symmetrized_structure()
    equivalent_positions = sym_structure.as_dict()['equivalent_positions']

    # build directed multi-graph
    twoway = nx.MultiDiGraph(graph.graph)
    # nodes are labelled by species
    nx.set_node_attributes(twoway,
        {node: {'symm_equiv': (structure[node].specie, equivalent_positions[node])} for node in twoway})

    def flip_edge(u, v, data):
        return v, u, {key: tuple(-i for i in val) if key == 'to_jimage'
                      else val for key, val in data.items()}

    # edges are labelled by periodic image crossing
    twoway.add_edges_from([flip_edge(*edge) for edge in twoway.edges(data=True)])

    return twoway


def split_molecular_graph(graph, filter_unique=False):

    connected = nx.connected_components(nx.Graph(graph))

    subgraphs = map(graph.subgraph, connected)

    def filter_unique_graphs(graphs):

        edge_match = iso.numerical_edge_match("weight", 1.0)
        node_match = iso.categorical_node_match("symm_equiv", None)

        unique_graphs = set()

        for graph in graphs:

            def graph_match(g):
                return nx.is_isomorphic(
                    graph, g, node_match=node_match, edge_match=edge_match)

            already_present = map(graph_match, unique_graphs)

            if not any(already_present):
                unique_graphs.add(graph)
                yield graph

    if filter_unique:
        return list(filter_unique_graphs(subgraphs))
    else:
        return list(subgraphs)


def extract_molecule(structure, graph, central_index=None):

    # Imports pymatgen modules
    p_structure = ModuleImporter('pymatgen.core.structure')
    Molecule = p_structure.Molecule

    # walk graph and consider to_jimage
    def generate_shifts():

        start = next(iter(graph.nodes())) if central_index is None else central_index

        def walk(shifts, edge):
            a, b = edge
            return shifts | {b: shifts[a] + graph[a][b][0]['to_jimage']}

        edges = nx.bfs_edges(graph, source=start)
        shifts = reduce(walk, edges, {start: np.zeros(3)})
        return dict(sorted(shifts.items()))

    shifts = generate_shifts()

    species = [structure.species[idx] for idx in shifts]
    coords = [structure.lattice.get_cartesian_coords(
        structure.frac_coords[idx] + shift) for idx, shift in shifts.items()]
    molecule = Molecule(species, coords)

    return list(shifts.keys()), molecule


def get_unique_entities(structure):

    # Imports pymatgen modules
    p_analyzer = ModuleImporter('pymatgen.symmetry.analyzer')
    SpacegroupAnalyzer = p_analyzer.SpacegroupAnalyzer

    sym_structure = SpacegroupAnalyzer(structure).get_symmetrized_structure()
    structure_dict = sym_structure.as_dict()
    equivalent_positions = structure_dict['equivalent_positions']

    molecular_graph = generate_molecular_graph(structure)
    connected_graphs = split_molecular_graph(molecular_graph, filter_unique=True)
    indices, molecules = zip(*map(lambda x: extract_molecule(structure, x), connected_graphs))

    elements = [site.specie.symbol for site in structure]

    mappings = [[equivalent_positions[idx] for idx in shift_dict]
                for shift_dict in indices]

    return equivalent_positions, elements, mappings, molecules


def generate_qm_region(idx, graph, structure):
    connected_graphs = split_molecular_graph(graph, filter_unique=False)
    central_graph = next(filter(lambda nodes: idx in nodes, connected_graphs))
    _, mol = extract_molecule(structure, central_graph, central_index=idx)
    central_site = structure.sites[idx]
    shift = np.array([central_site.x, central_site.y, central_site.z])
    return mol, shift


def build_cluster(poscar, central_idc=None, cluster_expansion=None,
                  cluster_cutoff=None):

    if cluster_expansion is not None:
        cluster = UnitSupercell.from_poscar(
            poscar, cluster_expansion, central_idc=central_idc)
    elif cluster_cutoff is not None:
        cluster = UnitSphere.from_poscar(
            poscar, cluster_cutoff, central_idc=central_idc)
    else:
        ValueError("Invalid cluster specification!")

    return cluster


class UnitCluster:
    """Base class for cluster of specific morphology made from unit cells.

    Parameters
    ----------
    lat_vecs : list
        Lattice vectors of unit cell as rows
    coords : list
        Coordinates of each atom in cell as fraction of cell parameters
    atom_data : list
        List of tuples containing atomic numbers, symbols, and original indices
    centre : list
        Coordinates of the centre of interest

    Attributes
    ----------
    cart_coords : list
        Cartesian coordinates of the expanded and shifted structure
    labels : list
        Labels for each atom in the final cluster, including boundary info.
    """
    def __init__(self, lat_vecs, coords, atom_data, centre, *args, **kwargs):

        self.lat_vecs = lat_vecs
        self.unit_coords = coords
        self.atom_data = atom_data
        # keep atom_numbers for compatibility with spin_phonon_suite
        self.atom_numbers = [ad[0] for ad in self.atom_data]

        # Periodically pre-shift unit cell to centre of interest, based on expansion parity
        translated_coords = self.recentre_unitcell(coords, centre, *args)
        
        # Symmetrise the translated cell by duplicating and pre-labeling all boundary atoms
        symm_points = self._symmetrise_cell(translated_coords)
        
        # Neatly print the padded cell, including boundary atoms
        self._print_padded_cell(symm_points)

        # Generate the full cluster, merging overlaps and creating final labels
        cart_coords, self.labels, atom_fractions = self.generate_cluster(symm_points, *args)

        # Apply final translation to move the geometric centre of the cluster to the origin
        final_shift_frac = self.get_final_shift_vector(*args)
        final_shift_cart = final_shift_frac @ self.lat_vecs
        self.cart_coords = cart_coords - final_shift_cart

        self.frac_coords = self.cart_coords @ la.inv(self.lat_vecs)
        self.n_atoms = len(self.cart_coords)
        self.n_cell = len(list(self.generate_cell_idc(*args)))
        
        a1, a2, a3 = self.lat_vecs
        self.unit_volume = np.dot(a1, np.cross(a2, a3))

        # The following is for Ewald summation, where 2pi is included in the summation
        self.recip_vecs = np.array([
            np.cross(a2, a3),
            np.cross(a3, a1),
            np.cross(a1, a2)
        ]) / self.unit_volume
        
        # Perform and print the atom fraction consistency check
        # This is that the number of atoms is equal to the N * Z, where N is
        # the number of unit cells, and Z is atoms per unit cell (taking into 
        # account atom fractions)
        self._check_and_print_consistency(self.labels, atom_fractions, *args)

    @classmethod
    def from_poscar(cls, poscar_name, *args, central_idc=None, centre=None):
        atoms = read_vasp(poscar_name)
        coords = atoms.scaled_positions
        
        atom_data = [(num, sym, i+1) for i, (num, sym) in enumerate(zip(atoms.numbers, atoms.get_chemical_symbols()))]

        if centre is not None:
            centre_frac = centre
        elif central_idc is not None:
            centre_frac = np.mean(coords[list(central_idc)], axis=0)
        else:
            centre_frac = np.zeros(3)

        return cls(atoms.cell, coords, atom_data, centre_frac, *args)
    
    def _symmetrise_cell(self, frac_coords):
        """
        Creates a 'padded' cell by duplicating all boundary atoms and pre-labeling them
        with a charge fraction.
        Returns a list of tuples: (coordinate, original_index, atom_fraction)
        """
        symm_points = []
        
        for i, coord in enumerate(frac_coords):
            # Pre-label the atom with its fraction.
            # After recentring, coords are in [0, 1), so only check for 0.0
            is_boundary = np.isclose(coord, 0.0)
            on_boundary_count = np.sum(is_boundary)
            atom_fraction = 1.0 / (2**on_boundary_count)
            
            symm_points.append((coord, i, atom_fraction))

            if on_boundary_count == 0:
                continue

            # Generate all 2^N-1 other permutations for N boundary dimensions
            for j in range(1, 1 << on_boundary_count):
                new_coord = np.copy(coord)
                offset_indices = np.where(is_boundary)[0]
                
                for k, offset_idx in enumerate(offset_indices):
                    if (j >> k) & 1:
                        # Create symmetric partner at 1.0 for each 0.0 boundary
                        new_coord[offset_idx] = 1.0
                
                symm_points.append((new_coord, i, atom_fraction))
                
        return symm_points

    def generate_cluster(self, symm_points, *args):
        """Builds cluster, merges overlaps, and assigns boundary-aware labels."""

        # 1. Tile the symmetrised/pre-labeled points to create the raw cluster
        all_points = []
        for cell_idc in self.generate_cell_idc(*args):
            for frac_coord, original_idx, atom_fraction in symm_points:
                cart_coord = (frac_coord + cell_idc) @ self.lat_vecs
                all_points.append((cart_coord, original_idx, atom_fraction))

        # 2. Group points by rounding coordinates to find overlaps
        unique_points = {}
        for cart_coord, original_idx, atom_fraction in all_points:
            key = tuple(np.round(cart_coord, 6))
            if key not in unique_points:
                # Store the original, higher-precision coordinate the first time a key is seen
                unique_points[key] = {'coord': cart_coord, 'merges': []}
            unique_points[key]['merges'].append((original_idx, atom_fraction))

        # 3. Create final list of atoms and labels by summing atom fractions
        final_cart_coords = []
        final_labels = []
        final_atom_fractions = []

        for key, data in unique_points.items():
            # Use the stored, higher-precision coordinate for the final output
            final_cart_coords.append(data['coord'])
            merged_points = data['merges']

            original_idx, base_fraction = merged_points[0]
            symbol = self.atom_data[original_idx][1]
            original_label_idx = self.atom_data[original_idx][2]

            # Sum the atom fractions of all merged points
            total_atom_fraction = sum(p[1] for p in merged_points)
            final_atom_fractions.append(total_atom_fraction)

            # Determine boundary suffix from the total atom fraction
            boundary_suffix = ""
            if np.isclose(total_atom_fraction, 1.0):
                boundary_suffix = ""  # Interior
            elif np.isclose(total_atom_fraction, 0.5):
                boundary_suffix = ".f"  # Face
            elif np.isclose(total_atom_fraction, 0.25):
                boundary_suffix = ".e"  # Edge
            elif np.isclose(total_atom_fraction, 0.125):
                boundary_suffix = ".c"  # Corner
            # Handle other boundary cases
            else:
                base_fraction_is_edge = any(np.isclose(p[1], 0.25) for p in merged_points)
                if np.isclose(total_atom_fraction, 0.75):
                    boundary_suffix = ".e34" if base_fraction_is_edge else ".c68"
                elif np.isclose(total_atom_fraction, 0.375):
                    boundary_suffix = ".c38"
                elif np.isclose(total_atom_fraction, 0.625):
                    boundary_suffix = ".c58"
                elif np.isclose(total_atom_fraction, 0.875):
                    boundary_suffix = ".c78"

            label = f"{symbol}.ENV{boundary_suffix}.{original_label_idx}"
            final_labels.append(label)

        return np.array(final_cart_coords), final_labels, final_atom_fractions

    def _print_padded_cell(self, symm_points):
        """Neatly prints the boundary atoms in the padded, pre-labeled unit cell."""
        print("\nBoundary atoms in translated unit cell:")
        print(f"{'Atom':<6} {'Orig Idx':<9} {'x':>8} {'y':>8} {'z':>8}   {'Atom Fraction':<15}")
        print("-" * 65)
        
        found_boundary_atom = False
        for coord, original_idx, atom_fraction in symm_points:
            if atom_fraction < 1.0:
                found_boundary_atom = True
                symbol = self.atom_data[original_idx][1]
                print(f"{symbol:<6} {original_idx:<9} {coord[0]:8.4f} {coord[1]:8.4f} {coord[2]:8.4f}   {atom_fraction:<15.4f}")
        
        if not found_boundary_atom:
            print("No boundary atoms found in this unit cell.")

        print("-" * 65 + "\n")

    def _check_and_print_consistency(self, final_labels, final_atom_fractions, *args):
        """Performs a final check and prints a summary of the generated cluster."""

        n_atoms_poscar = len(self.atom_data)
        n_tiled_cells = len(list(self.generate_cell_idc(*args)))
        total_atom_sum = sum(final_atom_fractions)
        
        # Map suffixes to descriptive names
        suffix_map = {
            "f": "Face", "e": "Edge", "c": "Corner",
            "e34": "Edge (3/4)", "c38": "Corner (3/8)", "c58": "Corner (5/8)",
            "c68": "Corner (6/8)", "c78": "Corner (7/8)"
        }

        # Count boundary types from the generated labels
        counts = Counter()
        for label in final_labels:
            if 'ENV' not in label:
                counts['QM'] += 1
                continue
            
            parts = label.split('.')
            if len(parts) >= 3 and parts[1] == 'ENV':
                if len(parts) == 3:
                    counts['Interior'] += 1
                else:
                    suffix = parts[2]
                    descriptive_name = suffix_map.get(suffix, f"Unknown ({suffix})")
                    counts[descriptive_name] += 1
            else:
                counts['Unknown'] += 1
        
        print("\nCluster consistency check:")
        print(f"{'Atoms in POSCAR:':<25} {n_atoms_poscar}")
        print(f"{'Tiled Unit Cells:':<25} {n_tiled_cells}")
        print("-" * 40)
        print(f"{'Total Atoms in Cluster:':<25} {self.n_atoms}")
        print("\nBoundary atom summary:")
        for b_type, count in sorted(counts.items()):
            print(f"  - {b_type+':':<22} {count}")
        print("-" * 40)
        print(f"{'Sum of atom fractions:':<25} {total_atom_sum:.4f}")
        
        # Corrected atom conservation check
        if np.isclose(total_atom_sum, n_tiled_cells * n_atoms_poscar):
            print("Atom conservation check:       PASS")
        else:
            print("Atom conservation check:       FAIL")
        print("-" * 40 + "\n")

    def recentre_unitcell(self, frac, centre, *args):
        """This method is dynamically handled in subclasses."""
        raise NotImplementedError("recentre_unitcell must be implemented in a subclass")

    def get_final_shift_vector(self, *args):
        """This method is dynamically handled in subclasses."""
        raise NotImplementedError("get_final_shift_vector must be implemented in a subclass")


class UnitSupercell(UnitCluster):

    def generate_cell_idc(self, expansion):
        """Generates symmetric cell indices for tiling."""
        def expansion_range(num):
            if num % 2 == 0:  # even, e.g., 2 -> [-1, 0]
                return range(-(num // 2), num // 2)
            else:  # odd, e.g., 3 -> [-1, 0, 1]
                return range(-(num // 2), num // 2 + 1)

        for nvec in product(*map(expansion_range, expansion)):
            yield nvec
            
    def recentre_unitcell(self, frac, centre, expansion):
        """
        Translates unit cell so the central atom is at the correct position
        (0.0 or 0.5) based on the parity of the supercell expansion.
        """
        # For even expansion, centre is 0.0. For odd, centre is 0.5.
        target_centre = np.array([0.0 if n % 2 == 0 else 0.5 for n in expansion])
        return (frac - centre + target_centre) % 1.0

    def get_final_shift_vector(self, expansion):
        """
        Calculates the vector needed to shift the geometric centre of the
        supercell to the origin.
        """
        return np.array([0.0 if n % 2 == 0 else 0.5 for n in expansion])


class UnitSphere(UnitCluster):

    def generate_cell_idc(self, cutoff):

        def expansion_range(vec1, vec2, vec3):
            norm_vec = np.cross(vec2, vec3)
            num = ceil(abs(cutoff / np.dot(norm_vec / la.norm(norm_vec), vec1)))
            return range(-num, num + 1)

        cyc_perm = [[self.lat_vecs[i - j] for i in range(3)] for j in range(3)]

        for nvec in product(*map(expansion_range, *cyc_perm)):

            r = np.sum([ni * ci for ni, ci in zip(nvec, self.lat_vecs)], axis=0)

            if la.norm(r) <= cutoff:
                yield nvec
    
    def recentre_unitcell(self, frac, centre, cutoff):
        """Translates unit cell so the central atom is at (0.5, 0.5, 0.5)."""
        return (frac - centre + 0.5) % 1.0

    def get_final_shift_vector(self, cutoff):
        """
        Calculates the vector needed to shift the geometric centre of the
        spherical cluster to the origin.
        """
        return np.array([0.5, 0.5, 0.5])


def write_molcas_basis_interior_only(labels, charge_dict, name):
    """
    Writes a Molcas basis file for environment charges,
    containing only interior (unscaled) charges.
    """
    with open(name, 'w') as f:
        f.write("* This file was generated by env_suite (interior charges only)\n")
        for elem, (lab, chrg) in zip(labels, charge_dict.items()):
            f.write(f"/{elem}.{name}.{lab}.0s.0s.\n")
            f.write("Dummy basis set for atomic charges of environment\n")
            f.write("no ref\n")
            f.write(f"{chrg:.9f} 0\n")
            f.write("0 0\n")
    return

def write_molcas_basis_specific(elements, charge_dict, name, required_types):
    """
    Writes a molcas basis file, containing only
    the necessary boundary types for each element.
    """
    
    # Map suffix to charge scaling factor
    scaling_factors = {
        "interior": 1.0, "f": 0.5, "e": 0.25, "c": 0.125,
        "c38": 0.375, "c58": 0.625, "e34": 0.75, "c68": 0.75, "c78": 0.875
    }
    
    with open(name, 'w') as f:
        f.write("* This file was generated by env_suite (including boundary charges)\n")
        
        # Iterate through all atoms in the POSCAR to handle non-equivalent sites
        for original_index, elem in enumerate(elements, start=1):
            if elem in required_types:
                charge = charge_dict[original_index]
                
                # Check which boundary types are actually present for this element
                # and write only those to the ENV file.
                for boundary_type in sorted(list(required_types[elem])):
                    suffix = "" if boundary_type == "interior" else f".{boundary_type}"
                    scaled_charge = charge * scaling_factors.get(boundary_type, 1.0)
                    
                    f.write(f"/{elem}.{name}{suffix}.{original_index}.0s.0s.\n")
                    f.write("Dummy basis set for atomic charges of environment\n")
                    f.write("no ref\n")
                    f.write(f"{scaled_charge:.9f} 0\n")
                    f.write("0 0\n")
    return


def write_basistype_table():
    """Writes the basistype.tbl file required by Molcas for ENV basis sets."""
    content = "ENV ANO AE_ RH_ PN_"
    with open("basistype.tbl", "w") as f:
        f.write(content)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: diable=line-too-long,missing-module-docstring,missing-function-docstring,missing-class-docstring,exec-used
import setuptools

# DO NOT EDIT THIS NUMBER
# It is changed automatically by python-semantic-release
__version__ = "0.8.0"

with open("README.md", "r") as file:
    long_description = file.read()

setuptools.setup(
        name="env_suite",
        version=__version__,
        author="Chilton Group",
        author_email="nicholas.chilton@manchester.ac.uk",
        description="A suite of tools for including enviroment effects in first principal calculations",
        long_description=long_description,
        long_description_content_type="text/markdown",
        project_url={
            "Bug Tracker": "https://github.com/chilton-group/env/issues",
            "Documentation": "https://chilton-group.gitlab.io/env_suite",
            "Source": "https://github.com/chilton-group/env_suite",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: OS Independent",
        ],
        packages=setuptools.find_packages(),
        python_requires=">=3.9",
        install_requires=[
            "numpy<=1.26.4",
            "scipy",
            "matplotlib",
            "hpc_suite>=1.9.0",
            "phonopy",
            "pymatgen",
            "gaussian_suite",
        ],
        entry_points={
            "console_scripts": [
                "env_suite=env_suite.cli:main"
                ]
            },
        )



"""
Command line interface for env_suite.
"""

import argparse
import re
from textwrap import dedent
import os
import numpy as np
from hpc_suite import parse_index
import gaussian_suite as gaussian
import xyz_py
import vasp_suite.structure as vasp

# from pymatgen.core.structure import Molecule, Structure

from . import ewald, plotter, utils, output
from . import cells


def ewald_func(args):
    """
    Wrapper function for ewald.

    Parameters
    ----------
    poscar : str
        Path to the POSCAR file.
    qm_region : str
        xyz file containing the QM region.
    charges : str
        .dat file containg the unit cell charges.
    central_atom : str
        The central atom of the QM region <Dy1>.
    expansion : list
        Supercell expansion <n n n>.
    r_cut : float
        Cutoff radius of the sphere region.
    n : list
        Summation cut off <n n n>.
    verbose : int
        Verbosity level.

    Returns
    -------
    ewald.out : file
        Output file containing charges to be entered into
        a muliconfiguration code.
    """
    # Import pymatgen
    p_structure = utils.ModuleImporter('pymatgen.core.structure')

    # Import class
    Structure = p_structure.Structure
    Molecule = p_structure.Molecule

    structure = Structure.from_file(args.poscar)
    central_idc = args.central_index

    if args.from_central:
        molecular_graph = cells.generate_molecular_graph(structure, bonding_cutoff=args.bonding_cutoff)
        qm_regions, shifts = zip(*map(lambda i: cells.generate_qm_region(i, molecular_graph, structure), central_idc))
        qm_region = Molecule.from_sites([site for mol in qm_regions for site in mol.sites])
        shift = np.mean(shifts, axis=0)
    elif args.qm_region is not None:
        qm_region = Molecule.from_file(args.qm_region)
        shift = np.zeros(3)
    elif args.coordination_sphere:
        poscar = vasp.Molecule(args.poscar, args.atom_index, scaler=1.1)
        poscar.get_coordination_sphere(args.coordination_sphere_number)
        coords = np.array(list(map(lambda x: x.coordinate, poscar.qm_graph.nodes)), dtype=float)
        atoms = np.array(list(map(lambda x: x.symbol, poscar.qm_graph.nodes)), dtype=str)
        atoms = list(map(str, atoms))
        shift = np.zeros(3)
    elif args.solid_state_dimer:
        poscar = vasp.Molecule(args.poscar, args.atom_index, args.max_bond_length)
        poscar.bridge_dimer()
        coords = np.array(list(map(lambda x: x.coordinate, poscar.qm_graph.nodes)), dtype=float)
        atoms = np.array(list(map(lambda x: x.symbol, poscar.qm_graph.nodes)), dtype=str)
        atoms = list(map(str, atoms))
        shift = np.zeros(3)
    else:
        raise NotImplementedError()

    if not args.coordination_sphere and not args.solid_state_dimer:
        coords = np.array([site.coords for site in qm_region.sites])
        atoms = [site.specie.symbol for site in qm_region.sites]

    coords -= shift
    utils.write_xyz('qm_region.xyz', coords, atoms)

    cluster = cells.build_cluster(args.poscar, central_idc=central_idc,
                                  cluster_expansion=args.cluster_expansion,
                                  cluster_cutoff=args.cluster_cutoff)

    calc_dict = ewald.ewald_potential(
            cluster=cluster,
            qm_region=coords,
            charges=args.charges,
            r_cut=args.r_cut,
            n=args.n,
            verbose=args.verbose)

    calc_dict.update({'qm atoms': atoms})

    h5out = output.OutputWriter('ewald.hdf5')
    h5out.write_h5(calc_dict)

    ewald.write_output(calc_dict['Sphere'],
                       calc_dict['Param'],
                       )


def plotter_func(args):
    """
    Wrapper function for plotter.

    Parameters
    ----------
    filename : str
        Path to the h5 file containing the charges.
    dim : int
        Dimension to plot the potential in /Å.
    num_points : int
        Number of points to evaluate the potential at.
    num_contours : int
        Number of contours to plot.
    show : bool
        Whether to show the plot or save it.

    Returns
    -------
    potential.png : file
        Potential plot.
    """
    if isinstance(args.num_points, list):
        num_points = args.num_points[0]
    else:
        num_points = args.num_points

    if isinstance(args.num_contours, list):
        num_contours = args.num_contours[0]
    else:
        num_contours = args.num_contours

    if not args.filename.split('.')[-1] == 'hdf5':
        raise ValueError('Input file must be a .hdf5 file')
    else:
        data = utils.parse_h5(args.filename)
        sphere, param = data['Sphere'], data['Param']
        coords = np.concatenate((sphere[:, :3], param[:, :3]), axis=0)
        q = np.concatenate((sphere[:, 4], param[:, 4]), axis=0)

    # evaluate the potential
    x, y, z = plotter.evaluate_potential(coords, q, num_points, args.dim)

    # plot the potential
    plotter.plot_potential(x, y, z, num_contours, args.show)


def plot_3d_model_func(args):
    """
    wrapper functions for plot_3d_model
    """
    if not args.filename.split('.')[-1] == 'hdf5':
        raise ValueError('Input file must be a .hdf5 file')
    else:
        data = utils.parse_h5(args.filename)
        sphere, param = data['Sphere'], data['Param']
        qm = data['qm region']

        sphere = np.array(sphere[:, :3], dtype=float)
        param = np.array(param[:, :3], dtype=float)

    plotter.plot_3d_model(sphere, param, qm)


def charges_func(args):
    """
    Wrapper function for cli call to charges
    """
    p_structure = utils.ModuleImporter('pymatgen.core.structure')
    Structure = p_structure.Structure
    structure = Structure.from_file(args.poscar)
    elements = [site.specie.symbol for site in structure]

    if args.gen:
        _, _, _, entities = cells.get_unique_entities(structure)
        gen_charges_func(entities, args)
    elif args.parse:
        equivalent_positions, _, mappings, entities = \
            cells.get_unique_entities(structure)
        parse_charges_func(equivalent_positions, elements, mappings, entities)
    elif args.from_cluster:
        charges_from_cluster_func(args)
    elif args.from_file:
        unitcell_charges = np.loadtxt(args.from_file)
        cells.write_molcas_basis_interior_only(
            elements, dict(enumerate(unitcell_charges, start=1)), "ENV")
        cells.write_basistype_table()
        print("\033[93m")
        print("Molcas basis set file (interior charges only) written to:")
        print("     ENV (basis set specification)\033[0m")
    else:
        ValueError("Invalid mode for 'charges' command.")


def charges_from_cluster_func(args):
    """
    Wrapper function for creating an ENV file including scaled boundary charges.
    """
    def _load_molcas_xyz(filename):
        """A simple local parser for xyz files with Molcas-style labels."""
        labels = []
        with open(filename, 'r') as f:
            lines = f.readlines()
            # Skip header lines (atom count and comment)
            for line in lines[2:]:
                parts = line.split()
                if len(parts) >= 4:
                    labels.append(parts[0])
        return labels

    # 1. Read the raw charges for each atom in the unit cell
    unitcell_charges = np.loadtxt(args.charges)
    
    # 2. Read the generated cluster.xyz file to see which atoms/boundaries exist
    labels = _load_molcas_xyz(args.from_cluster)

    # 3. Determine the unique set of boundary types needed for each element
    required_types = {}
    for label in labels:
        parts = label.split('.')
        if len(parts) < 2 or 'ENV' not in parts[1]:
            continue
        
        element = parts[0]
        if element not in required_types:
            required_types[element] = set()

        # Check for boundary suffix
        if len(parts) > 2 and parts[1] == 'ENV':
            if len(parts) == 3: # e.g., C.ENV.1
                required_types[element].add("interior")
            else: # e.g., C.ENV.f.1
                required_types[element].add(parts[2])
        else: 
            required_types[element].add("interior")

    # 4. Get element symbols from the POSCAR
    p_structure = utils.ModuleImporter('pymatgen.core.structure')
    Structure = p_structure.Structure
    structure = Structure.from_file(args.poscar)
    elements = [site.specie.symbol for site in structure]

    # 5. Write the ENV file including scaled boundary atom charges
    cells.write_molcas_basis_specific(
        elements, 
        dict(enumerate(unitcell_charges, start=1)), 
        "ENV", 
        required_types
    )
    cells.write_basistype_table()
    print("\033[93m")
    print("Molcas basis set file (including boundary charges) written to:")
    print("     ENV (basis set specification)\033[0m")


def format_formula(formula):

    def split_symbol_number(x):
        elem = re.search('([a-zA-Z]+)', x).group(0)
        num = int(re.search('(\d+)', x).group(0))
        return (elem, num)

    def format_constituent(x):
        return x[0] + str("" if x[1] == 1 else x[1])

    constituents = map(split_symbol_number, formula.split(" "))
    ordered = map(format_constituent, sorted(constituents, key=lambda x: x[0]))
    return ''.join(ordered)


def gen_charges_func(entities, args):
    """
    Wrapper function for cli call to charges gen
    """

    print("******************************")
    print("The unique entities are")
    for mol in entities:
        print(mol.formula)
        mol.to(f"{format_formula(mol.formula)}.xyz")
    print("******************************")

    # Create gaussian inputs for CHELPG calculation, one for each
    # member of the ASU

    # def2-SVP for elements: H-La, Hf-Rn 
    def2_svp =  ["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na",
                "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc", "Ti",
                "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge",
                "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo",
                "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te",
                "I", "Xe", "Cs", "Ba", "La", "Hf", "Ta", "W", "Re", "Os",
                "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn"]
    
    # def2 family is all electron H-Kr. Rb-Rn uses an ecp, 'Def2SVP' 
    # in the Gaussian basis library.
    def2_ecp = ["Rb", "Sr", "Y", "Zr", "Nb", "Mo",
                "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te",
                "I", "Xe", "Cs", "Ba", "La", "Hf", "Ta", "W", "Re", "Os",
                "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn"]

    def2_basis = {atom: "Def2SVP" for atom in def2_svp}

    # Stuttgart RSC 1997 ECP on Ln and An by default
    if args.stuttgart_ecp_3plus:
        pseudo = {metal: "Stuttgart-3plus"
                 for metal in xyz_py.atomic.f_block}
    else:
        pseudo = {metal: "stuttgart rsc 1997"
                 for metal in xyz_py.atomic.f_block}
    
    for atom in def2_ecp:
        pseudo[atom] = "Def2SVP"

    try:
        os.mkdir("gaussian")
    except OSError:
        pass

    for idx, mol in enumerate(entities, start=1):
        gaussian.gen_input.gen_input(
            f"gaussian/{format_formula(mol.formula)}_{idx}.com",
            list(map(lambda site: site.specie.symbol, mol.sites)),
            list(map(lambda site: (site.x, site.y, site.z), mol.sites)),
            99,
            99,
            method="PBE",
            bs_spec=def2_basis,
            ecp_spec=pseudo,
            opt=False,
            freq=False,
            extra_title=f"{format_formula(mol.formula)} (molecule #{idx})",
            chelpg="charge"
        )
        print()

    print("\033[93mCharge and multiplicity set to 99 in all .com files")
    print("Please replace with actual values before submitting \033[0m \n")

    return


def parse_charges_func(equivalent_positions, elements, mappings, entities):
    """
    Wrapper function for cli call to charges parse.
    This function reads charges from Gaussian logs, neutralizes them,
    writes them to charges.dat, and creates an ENV file with only
    interior charges.
    """

    def read_charges(mol, idx):
        file = f"gaussian/{format_formula(mol.formula)}_{idx}.log"
        charges = gaussian.cd_extractor.get_chelpg_charges(file)

        if not charges:
            raise ValueError(f"{file} does not contain charges!")

        return np.array(charges)

    def symmetrize(mapping, charges):
        for val in np.unique(mapping):
            for idc in np.flatnonzero(mapping == val):
                yield val, np.mean(charges[idc])

    def neutralize(charges):
        return np.array(charges) - np.mean(charges)

    mol_mappings = enumerate(zip(mappings, entities), start=1)
    atomic_charges = {idx: chrg for mol_idx, (mapping, mol) in mol_mappings
                      for idx, chrg in symmetrize(mapping, read_charges(mol, mol_idx))}

    unitcell_charges = neutralize([
        atomic_charges[pos] for pos in equivalent_positions])

    np.savetxt("charges.dat", unitcell_charges)
    print("Neutralized unit cell charges written to charges.dat")


    # Make a molcas basis file with only interior charges
    cells.write_molcas_basis_interior_only(
        elements, dict(enumerate(unitcell_charges, start=1)), "ENV")
    cells.write_basistype_table()
    print("\033[93m")
    print("Molcas basis set file (interior charges only) written to:")
    print("     ENV (basis set specification)\033[0m")


def cluster_func(args):

    p_structure = utils.ModuleImporter("pymatgen.core.structure")
    Structure = p_structure.Structure
    Molecule = p_structure.Molecule

    structure = Structure.from_file(args.poscar)
    central_idc = args.central_index

    if args.from_central:
        molecular_graph = cells.generate_molecular_graph(structure, bonding_cutoff=args.bonding_cutoff)
        qm_regions, shifts = zip(*map(lambda i: cells.generate_qm_region(i, molecular_graph, structure), central_idc))
        qm_region = Molecule.from_sites([site for mol in qm_regions for site in mol.sites])
        shift = np.mean(shifts, axis=0)
    elif args.qm_region is not None:
        qm_region = Molecule.from_file(args.qm_region)
        shift = np.zeros(3)
    elif args.coordination_sphere:
        # poscar = vasp.Molecule(args.poscar, args.atom_index, scaler=1.1)
        # poscar.get_coordination_sphere(args.coordination_sphere_number)
        # poscar.write_xyz("coordination_sphere.xyz")
        poscar = vasp.Molecule(args.poscar, args.atom_index, scaler=1.1)
        poscar.get_coordination_sphere(args.coordination_sphere_number)
        poscar.qm_graph.to_xyz("coordination_sphere.xyz", mass_centered=False)
        qm_region = Molecule.from_file("coordination_sphere.xyz")
        shift = np.zeros(3)
    else:
        raise NotImplementedError()

    cluster = cells.build_cluster(args.poscar, central_idc=central_idc,
                                  cluster_expansion=args.cluster_expansion,
                                  cluster_cutoff=args.cluster_cutoff)

    # Get the pre-generated labels from the cluster object
    labels = cluster.labels

    def generate_mapping(ref_coords):

        for idx, site in enumerate(qm_region.sites):

            coord = np.array([site.x, site.y, site.z]) - shift
            idc = np.flatnonzero(np.isclose(ref_coords, coord).all(axis=1))

            if len(idc) == 0:
                continue
            elif len(idc) == 1:
                yield idc[0]
            else:
                raise ValueError(f"Multiple instances of atom {coord}!")

    # Re-label the QM atoms to remove the .ENV suffix
    qm_indices = list(generate_mapping(cluster.cart_coords))
    for idx in qm_indices:
        labels[idx] = labels[idx].split('.')[0]

    xyz_py.save_xyz("cluster.xyz", labels, cluster.cart_coords)


cluster = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False
)

cluster_generation = cluster.add_argument_group("Cluster generation")
morph = cluster_generation.add_mutually_exclusive_group(required=True)

morph.add_argument(
    '--cluster_expansion',
    nargs=3,
    metavar=('N_x', 'N_y', 'N_z'),
    type=int,
    help='Supercell expansion. (Ewald)'
)

morph.add_argument(
    '--cluster_cutoff',
    type=float,
    help='Cut-off distance for unit cell cluster. (Reaction Field)'
)

cluster.add_argument(
    '--central_index',
    type=parse_index,
    nargs='+',
    help='Index of the central spin center or list of atomic indices.'
)

cluster.add_argument(
    '--bonding_cutoff',
    type=float,
    default=3.0,
    help='Cutoff distance for identifying molecular bonds (default: 3.0 Å).'
)


def get_index(atoms: list,
              atom: str,
              natoms: list,
              ) -> np.ndarray:

    atom = re.split(r'(\d+)', atom)
    for ind, sym in enumerate(atoms):
        if sym == atom[0]:
            index = ind

    prev = np.sum(natoms[:index])
    return int(prev) + int(atom[1]) - 1


def read_args(arg_list=None):
    '''Reads the command line arguments.'''
    parser = argparse.ArgumentParser(
            prog='env_suite',
            description=dedent(
                '''
                Available programmes:
                    env_suite ewald ...
                    env_suite cluster ...
                    env_suite charges ...
                    env_suite plot_potential ...
                    env_suite visualise_ewald ...
                '''),
            epilog=dedent(
                '''
                To display options for a specific programme, use:
                    env_suite <programme> -h
                '''),
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )

    subparsers = parser.add_subparsers(dest='prog')

    ewald_parser = subparsers.add_parser(
        'ewald',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[cluster],
        )

    ewald_parser.set_defaults(func=ewald_func)

    ewald_parser.add_argument(
        '--poscar',
        type=str,
        help='Path to the POSCAR file.',
        default='POSCAR',
        )

    ewald_qm_method_parser = ewald_parser.add_argument_group("qm_region generation method options")

    ewald_qm = ewald_qm_method_parser.add_mutually_exclusive_group(required=True)

    ewald_qm.add_argument(
        '--from_central',
        action='store_true',
        help=('QM region is built by completing the molecule around the central index. Not required with --qm_region')
    )

    ewald_qm.add_argument(
        '--qm_region',
        type=str,
        help='Coordinates of the QM-region.'
    )

    ewald_qm.add_argument(
            '--coordination_sphere',
            action='store_true',
            help='Build QM region from coordination sphere.'
    )

    ewald_qm.add_argument(
            '--solid_state_dimer',
            action='store_true',
            help='Build QM region from solid state dimer.'
    )

    qm_region_parser = ewald_parser.add_argument_group("--qm_region options")

    qm_region_parser.add_argument(
            '--atom_index',
            type=str,
            help='Atom index of the central atom. eg "Dy1"'
    )

    coordination_sphere_parser = ewald_parser.add_argument_group("--coordination_sphere options")

    coordination_sphere_parser.add_argument(
            '--coordination_sphere_number',
            type=int,
            help='Number of atoms in coordination sphere.'
    )

    ewald_parser.add_argument(
            '--charges',
            type=str,
            help='.dat file containg the unit cell charges.',
            )

    ewald_parser.add_argument(
            '--r_cut',
            type=float,
            help='Cutoff radius of the exact point charge region within the cluster_expansion',
            )

    ewald_parser.add_argument(
            '--n',
            type=int,
            nargs=3,
            help='Summation cut off.',
            metavar=('N_x', 'N_y', 'N_z'),
            default=[1, 1, 1],
            )

    ewald_parser.add_argument(
            '--verbose', '-v',
            type=int,
            help=dedent(
                '''
                Verbosity level
                0: No Output
                1: Summary of fitting
                2: Full fitting Output
                '''),
            default=1,
            )

    plotter_parser = subparsers.add_parser(
        'plot_potential',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    plotter_parser.set_defaults(func=plotter_func)

    plotter_parser.add_argument(
        '--filename', '-i',
        type=str,
        help='Path to the file containing the charges.',
        default='ewald.hdf5',
        )

    plotter_parser.add_argument(
        '--num_points', '-np',
        type=int,
        help='Number of points to evaluate the potential at.',
        default=50,
        nargs=1,
        )

    plotter_parser.add_argument(
            '--num_contours', '-nc',
            type=int,
            help='Number of contours to plot.',
            default=100,
            nargs=1,
            )

    plotter_parser.add_argument(
        '--show', '-s',
        action='store_true',
        help='Shows the plot',
        )

    plotter_parser.add_argument(
        '--dim', '-d',
        type=int,
        help='Dimension to plot the potential in /Å.',
        metavar='DIM_X, DIM_Y',
        default=[10, 10],
        nargs=2,
        )

    plot3d = subparsers.add_parser(
            "visualise_ewald",
            description="""
            Plots a 3d model of the ewald zones
            """)

    plot3d.set_defaults(func=plot_3d_model_func)

    plot3d.add_argument(
            '--filename', '-f',
            type=str,
            help='hdf5 file from ewald calculation',
            default='ewald.hdf5',
            )

    charges = subparsers.add_parser(
        "charges",
        description="""
        Creates inputs for Gaussian CHELPG charge calculations of each
        entity in the unit cell of a VASP optimised structure and
        collects the resulting charges
        """)

    charges.set_defaults(func=charges_func)

    charges.add_argument(
        "poscar",
        type=str,
        help='Poscar containing optimised geometry'
    )

    charges.add_argument(
        "--stuttgart_ecp_3plus",
        action='store_true',
        help='Use Stuttgart 3+ f-in-core ECPs on all Ln and An atoms when generating Gaussian inputs. By default Stuttgart RSC 1997 ECPs are used (these ECPs do NOT include f-electrons in the core).'
    )

    build = subparsers.add_parser('cluster', parents=[cluster])
    build.set_defaults(func=cluster_func)

    build.add_argument(
        '--poscar',
        type=str,
        help='Unit cell POSCAR.'
    )


    mode = charges.add_mutually_exclusive_group(required=True)

    mode.add_argument(
        "--gen",
        action='store_true',
        help='Generate Gaussian ChelpG inputs.'
    )
    
    mode.add_argument(
        "--parse",
        action='store_true',
        help='Parse Gaussian ChelpG outputs to create charges.dat and a ENV file containing only interior charges.'
    )
    
    mode.add_argument(
        "--from_file",
        type=str,
        metavar='CHARGES_DAT',
        help='Generate a ENV file from a charges.dat file containing only interior charges.'
    )

    mode.add_argument(
        "--from_cluster",
        type=str,
        metavar='CLUSTER_XYZ',
        help='Generate a cluster-specific ENV file from a cluster.xyz file, which contains appropriate fractional boundary charges.'
    )

    charges.add_argument(
        '--charges',
        type=str,
        help='Path to the charges.dat file. Required for --from_cluster.'
    )

    build_qm = build.add_argument_group("qm regions generation option")
    qm = build_qm.add_mutually_exclusive_group(required=True)

    qm.add_argument(
        '--from_central',
        action='store_true',
        help=('QM region is built by completing the molecule around the '
              'central index.')
    )

    qm.add_argument(
        '--qm_region',
        type=str,
        help='Coordinates of the QM-region.'
    )

    qm.add_argument(
        '--coordination_sphere',
        action='store_true',
        help='Build QM region from coordination sphere.'
    )

    build_qm_opt = build.add_argument_group("--qm_region options")
    build_qm_opt.add_argument(
        '--atom_index',
        type=str,
        help='Atom index of the central atom. eg "Dy1" required with --qm_region'
    )

    build_sphere = build.add_argument_group("--coordination_sphere options")
    build_sphere.add_argument(
        '--coordination_sphere_number',
        type=int,
        help='Number of atoms in coordination sphere.'
    )

    # Parse the arguments 
    parser.set_defaults(func=lambda args: parser.print_help())
    args = parser.parse_args(arg_list)
    
    if args.prog == 'charges' and args.from_cluster and not args.charges:
        parser.error("--charges is required when using --from_cluster")

    # Select programme
    if args in ['ewald', 'plot_potential']:
        args.func(args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)


def main():
    read_args()

